import unittest

from mock import MagicMock
from mock import patch

from py_i2c_register.register import Register
from py_i2c_register.register_segment import RegisterSegment

class TestRegisterInit(unittest.TestCase):
    def test_perfect(self):
        reg = Register("NAME", 1, 2, "WRITEMODE", {"key": "value"})

        self.assertEqual(reg.name, "NAME")
        self.assertEqual(reg.dev_addr, 1)
        self.assertEqual(reg.reg_addr, 2)
        self.assertEqual(reg.op_mode, "WRITEMODE")
        self.assertEqual(reg.segments, {"key": "value"})

class TestRegisterGet(unittest.TestCase):
    def setUp(self):
        self.reg = Register("NAME", 1, 2, "WRITEMODE", {})
        self.reg.add("SEG_NAME", 0, 2, [0] * 3)

    def test_perfect(self):
        ret = self.reg.get("SEG_NAME")
        self.assertEqual(ret.name, "SEG_NAME")
        self.assertEqual(ret.lsb_i, 0)
        self.assertEqual(ret.msb_i, 2)
        self.assertEqual(ret.bits, [0] * 3)

    def test_non_existant(self):
        with self.assertRaises(KeyError):
            self.reg.get("DOES_NOT_EXIST")

class TestRegisterProxyMethods(unittest.TestCase):
    def setUp(self):
        self.reg = Register("NAME", 1, 2, "WRITEMODE", {})
        self.reg.add("SEG_NAME", 0, 2, [0] * 3)

    def test_set_bits(self):
        self.reg.get("SEG_NAME").set_bits = MagicMock()
        self.reg.set_bits("SEG_NAME", [1, 0, 1])
        self.reg.get("SEG_NAME").set_bits.called_once_with([1, 0, 1])

    @patch("py_i2c_register.register_segment.RegisterSegment.num_bytes_for_bits")
    def test_len_bytes(self, fn):
        self.reg.len_bytes()
        fn.assert_called_once_with(len(self.reg))

class TestRegisterAdd(unittest.TestCase):
    def test_perfect(self):
        reg = Register("NAME", 1, 2, "WRITEMODE", {})
        reg.add("SEG_NAME", 0, 2, [0] * 3)

        ret = reg.get("SEG_NAME")
        self.assertEqual(ret.name, "SEG_NAME")
        self.assertEqual(ret.lsb_i, 0)
        self.assertEqual(ret.msb_i, 2)
        self.assertEqual(ret.bits, [0] * 3)

class TestRegisterRead(unittest.TestCase):
    def test_perfect(self):
        i2c = MagicMock()
        i2c.readBytes = MagicMock(return_value=[213, 170])

        reg = Register("NAME", 1, 2, Register.READ, {})
        reg.add("SEG_NAME", 0, 2, [0] * 3)

        reg.read(i2c)

        self.assertEqual(reg.get("SEG_NAME").bits, [1, 0, 1])

    def test_not_configured_to_read(self):
        i2c = MagicMock()

        reg = Register("NAME", 1, 2, Register.WRITE, {})
        reg.add("SEG_NAME", 0, 2, [0] * 3)

        with self.assertRaises(AttributeError):
            reg.read(i2c)

    def test_not_enough_bytes_read(self):
        i2c = MagicMock()
        i2c.readBytes = MagicMock(return_value=[32])

        reg = Register("NAME", 1, 2, Register.READ, {})
        reg.add("SEG_NAME", 9, 11, [0] * 3)

        with self.assertRaises(KeyError):
            reg.read(i2c)

    def test_i2c_read_fail(self):
        i2c = MagicMock()
        i2c.readBytes = MagicMock(side_effect=Exception("Exception"))

        reg = Register("NAME", 1, 2, Register.READ, {})
        reg.add("SEG_NAME", 0, 2, [0] * 3)

        with self.assertRaises(SystemError):
            reg.read(i2c)

class TestRegisterWrite(unittest.TestCase):
    def setUp(self):
        self.reg = Register("NAME", 1, 2, Register.WRITE, {})
        self.reg.add("SEG_NAME", 0, 2, [0] * 3)

        self.i2c = MagicMock()
        self.i2c.writeBytes = MagicMock(return_value=None)

    def test_perfect(self):
        self.reg.get("SEG_NAME").bits = [0, 1, 1]

        self.reg.write(self.i2c)
        self.i2c.writeBytes.assert_called_once_with(1, 2, [6])

    def test_not_setup_to_write(self):
        self.reg.op_mode = Register.READ

        with self.assertRaises(AttributeError):
            self.reg.write(self.i2c)

    def test_non_cont_seg_bits(self):
        self.reg.add("BAD_SEG", 9, 11, [0] * 3)

        with self.assertRaises(SyntaxError):
            self.reg.write(self.i2c)

    def test_multiple_segments_managing_same_bit(self):
        self.reg.add("BAD_SEG", 2, 5, [0] * 4)

        with self.assertRaises(KeyError):
            self.reg.write(self.i2c)

    def test_multiple_segments_managing_same_bit_more_than_one_bit(self):
        self.reg.add("BAD_SEG1", 2, 5, [0] * 4)
        self.reg.add("BAD_SEG2", 5, 7, [0] * 3)

        with self.assertRaises(KeyError):
            self.reg.write(self.i2c)

    def test_i2c_write_fail(self):
        self.i2c.writeBytes.return_value = 1

        with self.assertRaises(SystemError):
            self.reg.write(self.i2c)

class TestRegisterGenericMethods(unittest.TestCase):
    def test_str(self):
        reg = Register("NAME", 1, 2, "OP_MODE", {})
        reg.add("SEG_NAME", 0, 2, [0] * 3)

        self.assertEqual(str(reg), "Register<name=NAME, address=2, op_mode=OP_MODE, segments={\n    SEG_NAME=RegisterSegment<name=SEG_NAME, lsb_i=0, msb_i=2, bits=[0, 0, 0]>\n}>")

    def test_len(self):
        reg = Register("NAME", 1, 2, "OP_MODE", {})
        reg.add("SEG1", 0, 2, [0] * 3)
        reg.add("SEG2", 3, 5, [0] * 3)

        self.assertEqual(len(reg), 6)



import re
from py_i2c_register.register_segment import RegisterSegment

class Register():
    """Wrapper class around a register on an i2c accessible device
    Fields:
      - name(str): Name of register, key used to access it
      - reg_addr(int): Address of register
      - dev_addr(int): Address of i2c device
      - op_mode(int): Flags which specify which kind of data operations can take place on register
      - bits(int or Bit[]): Number of Register Bit objects or array of Register Bit objects
    """
    READ = "READ"
    _read_pattern = re.compile(".*{}.*".format(READ))
    WRITE = "WRITE"
    _write_pattern = re.compile(".*{}.*".format(WRITE))

    """Creates a Register instance
    Args: Same as Fields
    """
    def __init__(self, name, dev_addr, reg_addr, op_mode, segments={}):
        self.name = name
        self.dev_addr = dev_addr
        self.reg_addr = reg_addr
        self.op_mode = op_mode
        self.segments = segments

    """Get RegisterSegment by name
    Returns:
        - RegisterSegment: RegisterSegment with name provided
    
    Raises:
        - KeyError: If no RegisterSegment with name provided exists
    """
    def get(self, name):
        if name not in self.segments:
            raise KeyError("No segment found with name: \"{}\"".format(name))

        return self.segments[name]

    """Set bits of Segment with name provided
    Args:
        - name(str): Name of Segment to set bits of
        - bits(int[]): Bits to set
    
    Raises:
        - KeyError: If Segment with name provided does not exist
        - IndexError: If length of bits array is less than that defined by lsb_i and msb_i
        - ValueError: If an element of the provided bits array is not equal to 0 or 1
    """
    def set_bits(self, name, bits):
        self.get(name).set_bits(bits)

    """Add Register segment
    Args: Same as RegisterSegment.__init__
    Returns:
      - Register: The current Register, for chain adding of RegisterSegments to same Register
    """
    def add(self, name, lsb_i, msb_i, bits):
        self.segments[name] = RegisterSegment(name, lsb_i, msb_i, bits)
        return self

    """Reads register
    Args:
      - i2c(I2C Object): I2C object used to communicate with i2c system, see docs/i2c-object.md for more information
    
    Raises:
      - AttributeError: If register is not configured to read
      - KeyError: If a register segment requests a bit that was not read
      - SystemError: If the I2C Object fails to read
    """
    def read(self, i2c):
        if Register._read_pattern.match(self.op_mode) is not None:
            # Get number of bytes to read, will raise AssertionError if segments do not create round number of bytes
            bytes_count = self.len_bytes()
            read_bytes = []

            try:
                read_bytes = i2c.readBytes(self.dev_addr, self.reg_addr, bytes_count)
            except Exception as e:
                raise SystemError("Failed to read i2c: {}".format(e))

            # Loop through each byte read and map to elements in RegisterSegment bit arrays
            for segment in self.segments:
                self.segments[segment].update_bits(read_bytes)  # Raises KeyError if we didn't read enough bytes

        else:
            raise AttributeError("Register {} is not set up to allow read operations, op_mode: \"{}\"".format(self.name, self.op_mode))

    """Writes register
    Args:
      - i2c(I2C Object): I2C object used to communicate with i2c system, see docs/i2c-object.md for more information
      
    Returns:
      - Register: Self for chaining
      
    Raises:
      - AttributeError: If register is not set up to write
      - SyntaxError: If RegisterSegments are not configured to make a continuous series of bits
      - KeyError: If two RegisterSegments are configured to manage the same bit
      - SystemError: Failed to write i2c
    """
    def write(self, i2c):
        if Register._write_pattern.match(self.op_mode) is not None:
            bits = {}
            managing_segment = {}
            max_bit_i = 0

            for segment in self.segments:
                segment = self.segments[segment]

                for bit_i in range(len(segment.bits)):
                    actual_bit_i = bit_i + segment.lsb_i
                    # Record maximum bit index for continuous test later
                    if actual_bit_i > max_bit_i:
                        max_bit_i = actual_bit_i

                    # Mark which segment manages bit in case of configuration error
                    if actual_bit_i not in managing_segment:
                        managing_segment[actual_bit_i] = []
                    managing_segment[actual_bit_i].append(segment.name)

                    bits[bit_i + segment.lsb_i] = segment.bits[bit_i]

            # Check bits are continuous and only 1 segment controls each bit
            cont_check_err_is = []  # Indexes where bit array is not continuous
            managing_segment_check_err_is = []  # Indexes where more than 1 RegisterSegment is controlling a bit
            for bit_i in range(max_bit_i + 1):
                # Continuous check
                if bit_i not in bits:
                    cont_check_err_is.append(bit_i)

                # Managing segment check
                if (bit_i in managing_segment) and (len(managing_segment[bit_i]) > 1):
                    managing_segment_check_err_is.append(bit_i)

            # Raise errors if there where any
            if len(cont_check_err_is) > 0:
                raise SyntaxError("RegisterSegments are not configured to make a continuous series of bits, no values at indexes: {}".format(cont_check_err_is))

            if len(managing_segment_check_err_is) > 0:
                indexes_msg = ""

                for i in range(len(managing_segment_check_err_is)):
                    bit_i = managing_segment_check_err_is[i]
                    indexes_msg += "{} (competing segments: {})".format(bit_i, managing_segment[bit_i])

                    if i != len(managing_segment_check_err_is) - 1:
                        indexes_msg += ", "

                raise KeyError("More than one RegisterSegment is managing the following bit indexes: {}".format(indexes_msg))

            # Convert from bits map to bits array
            bits_arr = []
            for key in bits:
                bit = bits[key]
                bits_arr.append(bit)

            # Create bytes array
            bytes_arr = RegisterSegment.to_padded_byte_arr(bits_arr)

            # Write to i2c
            write_status = i2c.writeBytes(self.dev_addr, self.reg_addr, bytes_arr)

            if write_status == 1:
                raise SystemError("Failed to write to i2c")
        else:
            raise AttributeError("Register {} is not set up to allow write operations, op_mode: \"{}\"".format(self.name, self.op_mode))

    """String representation of Register
    Returns:
        - str: String representation of Register
    """
    def __str__(self):
        out = "Register<name={}, address={}, op_mode={}, segments={{\n".format(self.name, self.reg_addr, self.op_mode)

        for key in self.segments:
            segment = self.segments[key]
            out += "    {}={}\n".format(key, str(segment))

        out += "}>"
        return out

    """Length of register is bytes
    Returns:
        - int: The total number of bytes the register has, rounds up
    """
    def len_bytes(self):
        return RegisterSegment.num_bytes_for_bits(len(self))

    """Length of register in bits
    Returns:
        - int: The total number of bits the register has
    """
    def __len__(self):
        l = 0
        for segment in self.segments:
            l += len(self.segments[segment])

        return l



from py_i2c_register.register import Register
from py_i2c_register.register_segment import RegisterSegment

class RegisterList():
    """A wrapper class around multiple Registers
    Fields:
        - dev_addr(int): I2C address of device which registers are on
        - i2c(I2C Object): I2C Object used to read and write registers, see docs/i2c-object.md for details
        - registers(map<str, Register>): Map of Registers, Keys are Register names and values are the Registers themselves
    """

    """Creates a new RegisterList instance
    Args: Same as fields
    """
    def __init__(self, dev_addr, i2c, registers):
        self.dev_addr = dev_addr
        self.i2c = i2c
        self.registers = registers

    """Converts a RegisterSegment bit array into an integer
    Args:
        - reg_name(str): Name of Register which houses Segment
        - seg_name(str): Name of Segment in Register
        - read_first(bool): If True the register will be read first
        
    Returns:
        - int: Integer value of Segment in Register
        
    Raises:
        - KeyError: If Register or Segment name provided does not exist
    """
    def to_int(self, reg_name, seg_name, read_first=False):
        # Convert value and return
        return self.get(reg_name, read_first=read_first).get(seg_name).bytes_to_int()

    """Converts a RegisterSegment bit array into an integer using two's compliment
    Args:
        - reg_name(str): Name of Register which houses Segment
        - seg_name(str): Name of Segment in Register
        - read_first(bool): If True the register will be read first
        
    Returns:
        - int: Integer value of Segment in Register
        
    Raises:
        - KeyError: If Register or Segment name provided does not exist
    """
    def to_twos_comp_int(self, reg_name, seg_name, read_first=False):
        return self.get(reg_name, read_first=read_first).get(seg_name).bytes_to_twos_comp_int()

    """Sets bits of RegisterSegment
    Args:
        - reg_name(str): Name of Register which segment is in
        - seg_name(str): Name of Segment to set bits for
        - bits(int[]): Bits to set
        - write_after(bool): If True will write register after bits are set
        - write_fn: Function used to write if not None
        
    Raises:
        - KeyError: If Register or Segment with provided name does not exist
        - IndexError: If bits array is not length set by Segment lsb_i and msb_i
        - ValueError: If bits array contains element with values other than 0 or 1
    """
    def set_bits(self, reg_name, seg_name, bits, write_after=False, write_fn=None):
        self.get(reg_name).get(seg_name).set_bits(bits)

        if write_fn is None:
            write_fn = self.write

        if write_after:
            write_fn(reg_name)

    """Sets bits of RegisterSegment from integer value
    Args:
        - reg_name: Same as Register.set_bits
        - seg_name: Same as Register.set_bits
        - val(int): Integer value to set
        - write_after: Same as Register.set_bits
        - write_fn: Function used to write if not None
        
    Raises: 
        - KeyError: Same as Register.set_bits
        - ValueError: If value provided can not be represented in Segment size
    """
    def set_bits_from_int(self, reg_name, seg_name, val, write_after=False, write_fn=None):
        seg = self.get(reg_name).get(seg_name)
        bits = RegisterSegment.to_bits(val, len(seg))

        self.set_bits(reg_name, seg_name, bits, write_after=write_after, write_fn=write_fn)

    """Add Register to register list
    Args: Same arguments as Register.__init__
    Returns:
        - Register: Recently added register, for chaining
      
    Raises:
        - KeyError: If register with name already exists
    """
    def add(self, name, address, op_mode, segments):
        if name in self.registers:
            raise KeyError("RegisterSegment with name already exists: name: {}".format(name))

        self.registers[name] = Register(name, self.dev_addr, address, op_mode, segments)
        return self.registers[name]

    """Get register with name
    Args:
      - name(str): Register name
      - read_first(bool): If True will read the Register with provided name first
      
    Returns:
      - Register: Register with specified name
      
    Raises:
      - KeyError: If register with name does not exist
    """
    def get(self, name, read_first=False):
        if name not in self.registers:
            raise KeyError("Register with name \"{}\" not found".format(name))

        # Read first if asked
        if read_first:
            self.read(name)

        return self.registers[name]

    """Read register
    Args:
      - name(str): Name of register to read
      
    Returns:
      - Register: Register that was read, with updated values
      
    Raises:
      - KeyError: If register with provided name does not exist
      - AttributeError: If register is not set up to read
    """
    def read(self, name):
        return self.get(name).read(self.i2c)

    """Write register
    Args:
      - name(str): Name of register to write
      
    Returns:
      - Register: Register that was written, for chaining
      
    Raises:
      - KeyError: If register with provided name does not exist
      - AttributeError: If register is not set up to write
    """
    def write(self, name):
        return self.get(name).write(self.i2c)

    """String representation of RegisterList
    Returns:
        - str: String representation of RegisterList
    """
    def __str__(self):
        out = "RegisterList<device_address={}, registers={{\n".format(self.dev_addr)

        for k in self.registers:
            v = self.registers[k]
            # Indent output from Register.str
            v = str(v)
            v = v.split("\n")

            newV = ""
            for i in range(0, len(v)):
                # Don't indent first line
                if i != 0:
                    newV += "    "

                newV += v[i]

                # No newline on last line
                if i != len(v) - 1:
                    newV += "\n"

            out += "    {}={}\n".format(k, newV)

        out += "}>"
        return out



import math

class RegisterSegment():
    """Class which holds information about section of register
    Fields:
      - name(str): Name of segment
      - lsb_i(int): Index of LSB
      - msb_i(int): Index of MSB
      - bits(int[]): List of bits, each element of list is either 0 or 1
    """

    """Converts a given number to a bit array
    Args:
        - number(int): Number to convert, must be in range [0, 2^size]
        - size(int): Number of bits used to represent number
    
    Return:
        - int[]: Array of integers representing number
      
    Raises:
        - ValueError: If number can not be represented by number of bits specified in size
    """
    @staticmethod
    def to_bits(number, size):
        if number < 0 or number > math.pow(2, size):
            raise ValueError(
                "Number provided must be in range: [0, {}], was: {}".format(int(math.pow(2, size)), number))

        bits = []
        for bit in format(number, "0{}b".format(size)):
            bits.insert(0, int(bit))

        return bits

    """Converts an array of bits into an integer
    Args:
        - bits(int[]): Array of bits to convert into an integer
        
    Returns:
        - int: Bits in integer form
    """
    @staticmethod
    def to_int(bits):
        out = 0
        l = len(bits)
        for bit in reversed(bits):
            out = (out << 1) | bit

        return out

    """Converts an array of bits arranged in two's compliment form into an integer
    Args:
        - bits(int[]): Array of bits to convert into an integer
        
    Returns:
        - int: Bits in integer form
    """
    @staticmethod
    def to_twos_comp_int(bits):
        bits_str = ""

        for b in reversed(bits):
            bits_str += str(b)

        size = len(bits)
        v = int(bits_str, 2)

        if (v & (1 << (size - 1))) != 0:
            v = v - (1 << size)

        return v

    """Calculate the minimum number of bytes needed to store a given number of bits
    Divides by 8 and rounds up.
    
    Args:
        - bits(int): The number of bits
    
    Returns:
        - int: The minimum number of bytes required to store number of bits
    """
    @staticmethod
    def num_bytes_for_bits(bits):
        return int(math.ceil(float(bits) / 8.0))

    """Converts an array of bits into a padded bytes array
    This just splits the bits array into groups of 8. It then fills any space at the end of the last pair with 0s. 
    The pairs of 8 bits are then converted into integers, and returned as a byte array.
    
    Args:
        - bits(int[]): Bits to convert into padded byte array
    
    Returns:
        - int[]: Byte array representation of bits
    """
    @staticmethod
    def to_padded_byte_arr(bits):
        bytes = []
        byte_slice_lower = 0  # Increases by 8 for each byte

        # Determine how many bytes the provided bits are and loop that many times
        # Each loop "synthesizes" a new byte from the bits array
        for byte_i in range(RegisterSegment.num_bytes_for_bits(len(bits))):
            # Check that upper limit isn't too big
            byte_slice_upper = ((byte_i + 1) * 8) - 1  # The index we *want* to slice to
            to_pad = 0  # Used to keep track of how many 0s we need to pad the end of this byte
            if byte_slice_upper > len(bits) - 1:
                # Keep track of the fact that we need to pad the end of this byte with some 0s
                to_pad = byte_slice_upper - (len(bits) - 1)

                # Resize if index we wanted to slice to is too big
                byte_slice_upper = len(bits) - 1


            # Convert
            # Add 1 to byte_slice_upper because upper range of slice is not inclusive
            byte_slice = bits[byte_slice_lower:byte_slice_upper + 1]

            # Append padding
            if to_pad > 0:
                byte_slice.extend([0] * to_pad)

            byte = RegisterSegment.to_int(byte_slice)
            bytes.append(byte)

            # Add 8 to lower slice bound for next byte
            byte_slice_lower += 8

        return bytes

    """Creates a Register Segment instance
    Raises:
        - IndexError: If length of bits array is less than that defined by lsb_i and msb_i
        - ValueError: If lsb_i or msb_i is not in the range [0, 7] or lsb_i and msb_i are greater or less than each other 
                      in wrong way
    """
    def __init__(self, name, lsb_i, msb_i, bits):
        self.name = name

        # Sanity check LSB and MSB indexes
        if lsb_i > msb_i:
            raise ValueError("LSB index can not be greater than MSB index, lsb_i: {}, msb_i: {}".format(lsb_i, msb_i))

        self.lsb_i = lsb_i
        self.msb_i = msb_i

        # Bit array
        self.set_bits(bits)

    """Calls RegisterSegment.to_int on self.bits
    Returns:
        - int: Bits in integer form
    """
    def bytes_to_int(self):
        return RegisterSegment.to_int(self.bits)

    """Calls RegisterSegment.to_twos_comp_into on self.bits
    Returns:
        - int: Bits converted to integer form by reversing two's compliment
    """
    def bytes_to_twos_comp_int(self):
        return RegisterSegment.to_twos_comp_int(self.bits)

    """Update RegisterSegment bits from given bytes array
    The bytes array is assumed to be the complete data read off of a register. Thus the first byte's LSB will be 
    treated as index 0 when dealing with the lsb_i and msb_i of the RegisterSegment.
    
    Args:
        - bytes(int[]): Bytes array to update bits values from
        
    Raises:
        - KeyError: If provided bytes array does not contain enough bytes to fill RegisterSegment values
    """
    def update_bits(self, bytes):
        # Check that bytes array contains values inside lsb_i and msb_i range
        min_bytes = RegisterSegment.num_bytes_for_bits(self.msb_i + 1)
        if len(bytes) < min_bytes:
            raise KeyError("Provided bytes array does not contain enough bytes to fill MSB, bytes: {}, MSB index: {}, required bytes length: {}".format(bytes, self.msb_i, min_bytes))

        # Determine start and end byte by dividing lsb_i by 8 and rounding down
        start_byte = int(math.floor(float(self.lsb_i) / 8.0))
        end_byte = int(math.floor(float(self.msb_i) / 8.0))

        # Convert needed bytes into bits
        # Keys will be offset by start_byte
        needed_bytes_as_bits = []

        for byte_i in range(start_byte, end_byte + 1):
            byte = bytes[byte_i]
            converted_bits = RegisterSegment.to_bits(byte, 8)

            # Convert bits and check bits are all 0 or 1
            for i in range(len(converted_bits)):
                converted_bits[i] = int(converted_bits[i])

            needed_bytes_as_bits.append(converted_bits)

        # Loop through bits
        for bit_i in range(self.lsb_i, self.msb_i + 1):
            in_byte_i = int(math.floor(float(bit_i) / 8.0))
            bit_offset = (in_byte_i * 8)  # Used to figure out which bit in the byte we are in

            self.bits[bit_i - self.lsb_i] = needed_bytes_as_bits[in_byte_i - start_byte][bit_i - bit_offset]

    """Set Segment bits
    Runs some sanity checks on the new bits before setting them.
    
    Args:
        - bits(int[]): Array of bits to set
    
    Raises:
        - IndexError: If length of bits array is less than that defined by lsb_i and msb_i
        - ValueError: If an element of the provided bits array is not equal to 0 or 1
    """
    def set_bits(self, bits):
        if len(bits) != len(self):
            raise IndexError(
                "Default list must be size that specified by lsb_i and msb_i, was: {}, should be: {}".format(len(bits),
                                                                                                             len(self)))

        i = 0
        for bit in bits:
            if bit != 0 and bit != 1:
                raise ValueError("Bits can only have the integer values 0 or 1, was: {}, bit_i: {}".format(bit, i))

            i += 1

        self.bits = bits

    def __str__(self):
        return "RegisterSegment<name={}, lsb_i={}, msb_i={}, bits={}>".format(self.name, self.lsb_i, self.msb_i,
                                                                              self.bits)

    """Get length of register segment
    Uses lsb_i and msb_i to calculate.
    
    Returns:
        - int: Number of bits represented by RegisterSegment.
    """
    def __len__(self):
        return self.msb_i - self.lsb_i + 1




import unittest

from mock import patch
from py_i2c_register.register_segment import RegisterSegment

class TestRegisterSegmentToBits(unittest.TestCase):
    def test_pos_in_range(self):
        self.assertEqual(RegisterSegment.to_bits(12, 8), [0, 0, 1, 1, 0, 0, 0, 0])

    def test_pos_out_range(self):
        with self.assertRaises(ValueError):
            RegisterSegment.to_bits(300, 8)

    def test_neg(self):
        with self.assertRaises(ValueError):
            RegisterSegment.to_bits(-2, 8)

class TestRegisterSegmentToInt(unittest.TestCase):
    def test_number(self):
        self.assertEqual(RegisterSegment.to_int([0, 0, 1, 1, 0, 0, 0, 0]), 12)

class TestRegisterSegmentToTwosCompInt(unittest.TestCase):
    def test_neg_number(self):
        self.assertEqual(RegisterSegment.to_twos_comp_int([0, 1, 0, 1, 1, 1, 1, 1]), -6)

    def test_pos_number(self):
        self.assertEqual(RegisterSegment.to_twos_comp_int([0, 0, 1, 1, 0, 0, 0, 0]), 12)

class TestRegisterSegmentNumBytesForBits(unittest.TestCase):
    def test_lower_range(self):
        self.assertEqual(RegisterSegment.num_bytes_for_bits(2), 1)

    def test_upper_range(self):
        self.assertEqual(RegisterSegment.num_bytes_for_bits(7), 1)

    def test_simple_overflow(self):
        self.assertEqual(RegisterSegment.num_bytes_for_bits(9), 2)

    def test_zero(self):
        self.assertEqual(RegisterSegment.num_bytes_for_bits(0), 0)

class TestRegisterSegmentToPaddedByteArr(unittest.TestCase):
    def test_empty_bits(self):
        self.assertEqual(RegisterSegment.to_padded_byte_arr([]), [])

    def test_one_byte_perfect_fit(self):
        self.assertEqual(RegisterSegment.to_padded_byte_arr([0, 0, 1, 1, 0, 0, 0, 0]), [12])

    def test_one_byte_partial_fit(self):
        self.assertEqual(RegisterSegment.to_padded_byte_arr([1, 1, 1, 1]), [15])

    def test_two_bytes_partial_fit(self):
        self.assertEqual(RegisterSegment.to_padded_byte_arr([1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1]), [85, 10])

    def test_two_bytes_perfect_fit(self):
        self.assertEqual(RegisterSegment.to_padded_byte_arr([0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0]), [170, 85])

class TestRegisterSegmentInit(unittest.TestCase):
    def test_perfect(self):
        seg = RegisterSegment("NAME", 0, 7, [0] * 8)
        self.assertEqual(seg.name, "NAME")
        self.assertEqual(seg.lsb_i, 0)
        self.assertEqual(seg.msb_i, 7)
        self.assertEqual(seg.bits, [0] * 8)

    def test_lsb_higher_than_msb(self):
        with self.assertRaises(ValueError):
            seg = RegisterSegment("NAME", 7, 0, [0] * 8)

    def test_provided_bits_too_small(self):
        with self.assertRaises(IndexError):
            seg = RegisterSegment("NAME", 0, 7, [0])

class TestRegisterSegmentProxyMethods(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.bits = [1, 0, 1, 0, 1, 0, 1, 0]

    def setUp(self):
        self.seg = RegisterSegment("NAME", 0, 7, self.bits)

    @patch("py_i2c_register.register_segment.RegisterSegment")
    def test_bytes_to_int(self, RS):
        self.seg.bytes_to_int()
        RS.to_int.assert_called_once_with(self.bits)

    @patch("py_i2c_register.register_segment.RegisterSegment")
    def test_bytes_to_twos_comp_int(self, RS):
        self.seg.bytes_to_twos_comp_int()
        RS.to_twos_comp_int.assert_called_once_with(self.bits)

class TestRegisterSegmentUpdateBits(unittest.TestCase):
    def test_perfect(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        seg.update_bits([213, 170])
        self.assertEqual(seg.bits, [1, 0, 1])

    def test_in_second_byte(self):
        seg = RegisterSegment("NAME", 9, 11, [0] * 3)
        seg.update_bits([213, 170])
        self.assertEqual(seg.bits, [1, 0, 1])

    def test_not_enough_bits(self):
        seg = RegisterSegment("NAME", 9, 11, [0] * 3)
        with self.assertRaises(KeyError):
            seg.update_bits([240])

class TestRegisterSegmentSetBits(unittest.TestCase):
    def test_perfect(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        seg.set_bits([1, 1, 0])
        self.assertEqual(seg.bits, [1, 1, 0])

    def test_too_many_bits(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        with self.assertRaises(IndexError):
            seg.set_bits([0, 1, 0, 1])

    def test_too_few_bits(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        with self.assertRaises(IndexError):
            seg.set_bits([0])

    def test_bit_value_err(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        with self.assertRaises(ValueError):
            seg.set_bits([1, 2, 3])

class TestRegisterSegmentGenericMethods(unittest.TestCase):
    def test_str(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        self.assertEqual(str(seg), "RegisterSegment<name=NAME, lsb_i=0, msb_i=2, bits=[0, 0, 0]>")

    def test_len(self):
        seg = RegisterSegment("NAME", 0, 2, [0] * 3)
        self.assertEqual(len(seg), 3)



import unittest
from mock import MagicMock

from py_i2c_register.register_list import RegisterList
from py_i2c_register.register import Register
from py_i2c_register.register_segment import RegisterSegment

class TestRegisterListInit(unittest.TestCase):
    def test_perfect(self):
        i2c = MagicMock()
        list = RegisterList(1, i2c, {"key": "value"})

        self.assertEqual(list.dev_addr, 1)
        self.assertEqual(list.i2c, i2c)
        self.assertEqual(list.registers, {"key": "value"})

class TestRegisterListProxyMethods(unittest.TestCase):
    def setUp(self):
        self.i2c = MagicMock()
        self.i2c.readBytes = MagicMock(return_value=[170])
        self.i2c.writeBytes = MagicMock()

        self.lst = RegisterList(1, self.i2c, {})
        self.lst.add("REG1", 1, Register.READ + Register.WRITE, {})
        self.lst.get("REG1").add("SEG1", 0, 2, [0] * 3)

    def test_to_int_read_first(self):
        self.lst.get("REG1").get("SEG1").bytes_to_int = MagicMock()

        self.lst.to_int("REG1", "SEG1", read_first=True)

        self.lst.get("REG1").get("SEG1").bytes_to_int.assert_called_once()
        self.i2c.readBytes.assert_called_once_with(1, 1, 1)

    def test_to_int_dont_read_first(self):
        self.lst.get("REG1").get("SEG1").bytes_to_int = MagicMock()

        self.lst.to_int("REG1", "SEG1", read_first=False)

        self.lst.get("REG1").get("SEG1").bytes_to_int.assert_called_once()
        self.i2c.readBytes.assert_not_called()

    def test_to_int_keyerror_reg(self):
        with self.assertRaises(KeyError):
            self.lst.to_int("DOES_NOT_EXIST", "SEG1")

    def test_to_int_keyerror_seg(self):
        with self.assertRaises(KeyError):
            self.lst.to_int("REG1", "DOES_NOT_EXIST")

    def test_to_twos_comp_int_read_first(self):
        self.lst.get("REG1").get("SEG1").bytes_to_twos_comp_int = MagicMock()

        self.lst.to_twos_comp_int("REG1", "SEG1", read_first=True)

        self.lst.get("REG1").get("SEG1").bytes_to_twos_comp_int.assert_called_once()
        self.i2c.readBytes.assert_called_once_with(1, 1, 1)

    def test_to_twos_comp_int_dont_read_first(self):
        self.lst.get("REG1").get("SEG1").bytes_to_twos_comp_int = MagicMock()

        self.lst.to_twos_comp_int("REG1", "SEG1", read_first=False)

        self.lst.get("REG1").get("SEG1").bytes_to_twos_comp_int.assert_called_once()
        self.i2c.readBytes.assert_not_called()

    def test_to_twos_comp_int_keyerror_reg(self):
        with self.assertRaises(KeyError):
            self.lst.to_twos_comp_int("DOES_NOT_EXIST", "SEG1")

    def test_to_twos_comp_int_keyerror_seg(self):
        with self.assertRaises(KeyError):
            self.lst.to_twos_comp_int("REG1", "DOES_NOT_EXIST")

    def test_set_bits_perfect_write_after(self):
        seg1 = self.lst.get("REG1").get("SEG1")
        seg1.set_bits = MagicMock(side_effect=seg1.set_bits)

        self.lst.set_bits("REG1", "SEG1", [1, 1, 0], write_after=True)

        seg1.set_bits.assert_called_once_with([1, 1, 0])
        self.i2c.writeBytes.assert_called_once_with(1, 1, [3])

    def test_set_bits_perfect_dont_write_after(self):
        seg1 = self.lst.get("REG1").get("SEG1")
        seg1.set_bits = MagicMock(side_effect=seg1.set_bits)

        self.lst.set_bits("REG1", "SEG1", [1, 1, 0], write_after=False)

        seg1.set_bits.assert_called_once_with([1, 1, 0])
        self.i2c.writeBytes.assert_not_called()

    def test_set_bits_perfect_write_after_custom_write_fn(self):
        seg1 = self.lst.get("REG1").get("SEG1")
        seg1.set_bits = MagicMock(side_effect=seg1.set_bits)

        mock_write = MagicMock()
        self.lst.set_bits("REG1", "SEG1", [1, 1, 0], write_after=True, write_fn=mock_write)

        seg1.set_bits.assert_called_once_with([1, 1, 0])
        mock_write.assert_called_once_with("REG1")

    def test_set_bits_perfect_dont_write_after_custom_write_fn(self):
        seg1 = self.lst.get("REG1").get("SEG1")
        seg1.set_bits = MagicMock(side_effect=seg1.set_bits)

        mock_write = MagicMock()
        self.lst.set_bits("REG1", "SEG1", [1, 1, 0], write_after=False, write_fn=mock_write)

        seg1.set_bits.assert_called_once_with([1, 1, 0])
        mock_write.assert_not_called()

    def test_set_bits_from_int(self):
        self.lst.set_bits = MagicMock()

        mock_write = MagicMock()
        self.lst.set_bits_from_int("REG1", "SEG1", 3, write_after=True, write_fn=mock_write)

        self.lst.set_bits.assert_called_once_with("REG1", "SEG1", [1, 1, 0], write_after=True, write_fn=mock_write)

    def test_read(self):
        reg1 = self.lst.get("REG1")
        reg1.read = MagicMock(side_effect=reg1.read)

        self.lst.read("REG1")

        reg1.read.assert_called_once_with(self.i2c)
        self.assertEqual(self.lst.to_int("REG1", "SEG1"), 2)

    def test_write(self):
        reg1 = self.lst.get("REG1")
        reg1.write = MagicMock(side_effect=reg1.write)

        self.lst.set_bits("REG1", "SEG1", [1, 1, 0])
        self.lst.write("REG1")

        reg1.write.assert_called_once_with(self.i2c)
        self.i2c.writeBytes.assert_called_once_with(1, 1, [3])

class TestRegisterListAdd(unittest.TestCase):
    def setUp(self):
        self.i2c = MagicMock()

        self.lst = RegisterList(1, self.i2c, {})

    def test_add_perfect(self):
       self.lst.add("REG1", 1, "OP_MODE", {"key": "value"})

       reg = self.lst.get("REG1")
       self.assertEqual(reg.name, "REG1")
       self.assertEqual(reg.dev_addr, 1)
       self.assertEqual(reg.op_mode, "OP_MODE")
       self.assertEquals(reg.segments, {"key": "value"})

    def test_add_already_exists(self):
        self.lst.add("REG1", 1, Register.READ, {})

        with self.assertRaises(KeyError):
            self.lst.add("REG1", 1, Register.READ, {})

class TestRegisterListGet(unittest.TestCase):
    def setUp(self):
        self.i2c = MagicMock()
        self.i2c.readBytes = MagicMock(return_value=[213])

        self.lst = RegisterList(1, self.i2c, {})
        self.seg1 = RegisterSegment("SEG1", 0, 2, [0] * 3)
        self.lst.add("REG1", 1, Register.READ, {"SEG1": self.seg1})

    def test_perfect_read_first(self):
        reg = self.lst.get("REG1", read_first=True)
        self.assertEqual(reg.name, "REG1")
        self.assertEqual(reg.dev_addr, 1)
        self.assertEqual(reg.op_mode, Register.READ)

        self.assertEquals(reg.segments["SEG1"], self.seg1)

        self.i2c.readBytes.assert_called_once_with(1, 1, 1)
        self.assertEqual(self.lst.get("REG1").get("SEG1").bits, [1, 0, 1])

    def test_perfect_dont_read_first(self):
        reg = self.lst.get("REG1", read_first=False)
        self.assertEqual(reg.name, "REG1")
        self.assertEqual(reg.dev_addr, 1)
        self.assertEqual(reg.op_mode, Register.READ)

        self.assertEqual(reg.segments["SEG1"], self.seg1)

        self.i2c.readBytes.assert_not_called()
        self.assertEqual(self.lst.get("REG1").get("SEG1").bits, [0, 0, 0])

    def test_keyerror(self):
        with self.assertRaises(KeyError):
            self.lst.get("DOES_NOT_EXIST")

class TestRegisterListGenericMethods(unittest.TestCase):
    def test_str(self):
        i2c = MagicMock()
        lst = RegisterList(1, i2c, {})
        lst.add("REG1", 1, Register.READ, {})\
            .add("SEG1", 0, 2, [0] * 3)

        self.assertEqual(str(lst), "RegisterList<device_address=1, registers={\n    REG1=Register<name=REG1, address=1, op_mode=READ, segments={\n        SEG1=RegisterSegment<name=SEG1, lsb_i=0, msb_i=2, bits=[0, 0, 0]>\n    }>\n}>")



from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

try:
    import pypandoc
    long_description = pypandoc.convert(path.join(here, "README.md"), 'rst')
except(IOError, ImportError) as e:
    raise Exception("Failed to convert README.md to rst format, error: {}".format(e))

setup(
    name="py-i2c-register",
    version="0.0.3",
    description="Python wrapper library around the common I2C controller register pattern.",
    long_description=long_description,
    url="https://github.com/Noah-Huppert/py-i2c-register",
    author="Noah Huppert",
    author_email="developer.noah@gmail.com",
    license="MIT",

    # List of: https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Project maturity
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",

        # Intended audience
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",

        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],

    keywords="library i2c registers",

    packages=find_packages(exclude=["contrib", "docs", "tests"]),

    install_requires=["pypandoc"],
    package_data={},
    data_files=[
        ("", ["README.md"])
    ],
    entry_points={}
)
